from unittest import TestCase
from unittest.mock import patch, Mock, MagicMock

from bauh import __app_name__
from bauh.api.abstract.controller import UpgradeRequirement
from bauh.gems.arch.dependencies import DependenciesAnalyser
from bauh.gems.arch.model import ArchPackage
from bauh.gems.arch.updates import UpdatesSummarizer
from bauh.view.util.translation import I18n


class UpdatesSummarizerGetUpgradeRequirementsTest(TestCase):

    def setUp(self) -> None:
        self.aur_client = MagicMock()
        self.aur_client.read_index.return_value = set()
        self.deps_analyser = MagicMock()
        self._i18n = {"arch.info.conflicts with": "", "arch.update_summary.to_install.dep_conflict": "{} {}",
                      "arch.info.required by": "", "arch.info.depends on": ""}
        self.i18n = I18n(default_key='en_US', default_locale=self._i18n, current_key='en_US', current_locale=self._i18n)
        self.summarizer = UpdatesSummarizer(aur_client=self.aur_client,
                                            aur_supported=True,
                                            i18n=self.i18n,
                                            logger=Mock(),
                                            deps_analyser=self.deps_analyser,
                                            watcher=Mock())
        self.config_ = {"automatch_providers": True,
                        "prefer_repository_provider": True,
                        "check_dependency_breakage": True}

    @patch(f"{__app_name__}.gems.arch.updates.pacman")
    def test__should_not_return_installed_to_remove_when_conflict_with_installed_version_fails(self, pacman: Mock):
        """
        If the newest version o package A conflicts with an installed package named B, but the conflict expression
        doesn't match (e.g: A conflict expressions (B <= 0.7), B (0.8)) then B should not be marked as a package
        to be removed.
        """
        pkg_a = ArchPackage(name="A", version="1.0.0-1", latest_version="1.1.0-1", repository="community")

        pacman.map_provided.side_effect = [{"A": {"A"},  # remote provided
                                            "A=1.1.0": {"A"},
                                            "B": {"B"},
                                            "B=0.8": {"B"}
                                            },
                                           {"A": {"A"},  # provided
                                            "A=1.0.0": {"A"},
                                            "B": {"B"},
                                            "B=0.8": {"B"}
                                            }
                                           ]
        pacman.map_repositories.return_value = {"A": pkg_a.repository, "B": "community"}
        pacman.map_updates_data.return_value = {"A": {'ds': 1,
                                                      's': 1,
                                                      'v': pkg_a.latest_version,
                                                      'c': {"B<=0.7"},
                                                      'p': {"A": {"A"},
                                                            "A=1.1.0": {"A"}},
                                                      'd': set(),
                                                      'r': "community",
                                                      'des': pkg_a.name}}
        pacman.map_installed.return_value = {"A": pkg_a.version, "B": "0.8-1"}
        pacman.get_installed_size.return_value = {"A": 1, "B": 1}
        pacman.map_required_by.return_value = {"A": set()}

        self.deps_analyser.map_missing_deps.return_value = list()

        res = self.summarizer.summarize(pkgs=[pkg_a], root_password=None, arch_config=self.config_)
        for method in ('map_provided', 'map_repositories', 'map_updates_data', 'map_installed',
                       'get_installed_size', 'map_required_by'):
            getattr(pacman, method).assert_called()

        for method in ('map_missing_deps',):
            getattr(self.deps_analyser, method).assert_called()

        self.assertFalse(res.to_remove)
        self.assertFalse(res.to_install)
        self.assertEqual([UpgradeRequirement(pkg=pkg_a, required_size=1, extra_size=0)], res.to_upgrade)

    @patch(f"{__app_name__}.gems.arch.updates.pacman")
    def test__should_return_installed_to_remove_when_conflict_with_installed_version_matches(self, pacman: Mock):
        """
        If the newest version o package A conflicts with an installed package named B and the conflict expression
        matches (e.g: A conflict expressions (B <= 0.7), B (0.7)) then B should be marked as a package to be removed.
        """
        pkg_a = ArchPackage(name="A", version="1.0.0-1", latest_version="1.1.0-1", repository="community")

        pacman.map_provided.side_effect = [{"A": {"A"},  # remote provided
                                            "A==1.0.1": {"A"},
                                            "B": {"B"},
                                            "B=0.7": {"B"}
                                            },
                                           {"A": {"A"},  # provided
                                            "A=1.0.0": {"A"},
                                            "B": {"B"},
                                            "B=0.7": {"B"}
                                            },
                                           {"B": {"B"},  # provided to remove
                                            "B=0.7": {"B"}
                                            }
                                           ]
        pacman.map_repositories.return_value = {"A": pkg_a.repository, "B": "community"}
        pacman.map_updates_data.return_value = {"A": {'ds': 1,
                                                      's': 1,
                                                      'v': pkg_a.latest_version,
                                                      'c': {"B<=0.7"},
                                                      'p': {"A": {"A"},
                                                            "A=1.1.0": {"A"}},
                                                      'd': set(),
                                                      'r': "community",
                                                      'des': pkg_a.name}}
        pacman.map_installed.return_value = {"A": pkg_a.version, "B": "0.7-1"}
        pacman.get_installed_size.return_value = {"A": 1, "B": 1}
        pacman.map_required_by.return_value = {"A": set()}

        self.deps_analyser.map_missing_deps.return_value = list()
        self.deps_analyser.map_all_required_by.return_value = set()

        res = self.summarizer.summarize(pkgs=[pkg_a], root_password=None, arch_config=self.config_)

        for method in ('map_provided', 'map_repositories', 'map_updates_data', 'map_installed',
                       'get_installed_size', 'map_required_by'):
            getattr(pacman, method).assert_called()

        for method in ('map_missing_deps', 'map_all_required_by'):
            getattr(self.deps_analyser, method).assert_called()

        self.assertFalse(res.to_install)
        self.assertEqual([UpgradeRequirement(pkg=pkg_a, required_size=1, extra_size=0)], res.to_upgrade)

        pkg_b = ArchPackage(name="B", installed=True, i18n=self.i18n)
        self.assertEqual([UpgradeRequirement(pkg=pkg_b, reason=" 'A'", extra_size=1)], res.to_remove)

    @patch(f"{__app_name__}.gems.arch.updates.pacman")
    def test__should_return_installed_to_remove_when_conflict_with_installed_matches(self, pacman: Mock):
        """
        If the newest version o package A conflicts with an installed package named B,
        then B should be marked as a package to be removed.
        """
        pkg_a = ArchPackage(name="A", version="1.0.0-1", latest_version="1.1.0-1", repository="community")

        pacman.map_provided.side_effect = [{"A": {"A"},  # remote provided
                                            "A==1.0.1": {"A"},
                                            "B": {"B"},
                                            "B=0.7": {"B"}
                                            },
                                           {"A": {"A"},  # provided
                                            "A=1.0.0": {"A"},
                                            "B": {"B"},
                                            "B=0.7": {"B"}
                                            },
                                           {"B": {"B"},  # provided to remove
                                            "B=0.7": {"B"}
                                            }
                                           ]
        pacman.map_repositories.return_value = {"A": pkg_a.repository, "B": "community"}
        pacman.map_updates_data.return_value = {"A": {'ds': 1,
                                                      's': 1,
                                                      'v': pkg_a.latest_version,
                                                      'c': {"B"},
                                                      'p': {"A": {"A"},
                                                            "A=1.1.0": {"A"}},
                                                      'd': set(),
                                                      'r': "community",
                                                      'des': pkg_a.name}}
        pacman.map_installed.return_value = {"A": pkg_a.version, "B": "0.7-1"}
        pacman.get_installed_size.return_value = {"A": 1, "B": 1}
        pacman.map_required_by.return_value = {"A": set()}

        self.deps_analyser.map_missing_deps.return_value = list()
        self.deps_analyser.map_all_required_by.return_value = set()

        res = self.summarizer.summarize(pkgs=[pkg_a], root_password=None, arch_config=self.config_)

        for method in ('map_provided', 'map_repositories', 'map_updates_data', 'map_installed',
                       'get_installed_size', 'map_required_by'):
            getattr(pacman, method).assert_called()

        for method in ('map_missing_deps', 'map_all_required_by'):
            getattr(self.deps_analyser, method).assert_called()

        self.assertFalse(res.to_install)
        self.assertEqual([UpgradeRequirement(pkg=pkg_a, required_size=1, extra_size=0)], res.to_upgrade)

        pkg_b = ArchPackage(name="B", installed=True, i18n=self.i18n)
        self.assertEqual([UpgradeRequirement(pkg=pkg_b, reason=" 'A'", extra_size=1)], res.to_remove)

    @patch(f"{__app_name__}.gems.arch.updates.pacman")
    def test__should_not_return_installed_to_remove_when_conflict_with_is_self_conflict(self, pacman: Mock):
        """
        If the newest version o package A conflicts with itself, then A should not be marked as a package to be removed.
        """
        pkg_a = ArchPackage(name="A", version="1.0.0-1", latest_version="1.1.0-1", repository="community")

        pacman.map_provided.side_effect = [{"A": {"A"},  # remote provided
                                            "A=1.1.0": {"A"},
                                            "C": {"A"}
                                            },
                                           {"A": {"A"},  # provided
                                            "A=1.0.0": {"A"},
                                            "C": {"A"}
                                            }
                                           ]
        pacman.map_repositories.return_value = {"A": pkg_a.repository}
        pacman.map_updates_data.return_value = {"A": {'ds': 1,
                                                      's': 1,
                                                      'v': pkg_a.latest_version,
                                                      'c': {"A<=1.0.0"},
                                                      'p': {"A": {"A"},
                                                            "A=1.1.0": {"A"},
                                                            "C": {"A"}},
                                                      'd': set(),
                                                      'r': "community",
                                                      'des': pkg_a.name}}
        pacman.map_installed.return_value = {"A": pkg_a.version}
        pacman.get_installed_size.return_value = {"A": 1}
        pacman.map_required_by.return_value = {"A": set()}

        self.deps_analyser.map_missing_deps.return_value = list()

        res = self.summarizer.summarize(pkgs=[pkg_a], root_password=None, arch_config=self.config_)
        for method in ('map_provided', 'map_repositories', 'map_updates_data', 'map_installed',
                       'get_installed_size', 'map_required_by'):
            getattr(pacman, method).assert_called()

        self.assertFalse(res.to_remove)
        self.assertFalse(res.to_install)
        self.assertEqual([UpgradeRequirement(pkg=pkg_a, required_size=1, extra_size=0)], res.to_upgrade)

    @patch(f"{__app_name__}.gems.arch.updates.pacman")
    def test__should_not_return_installed_to_remove_when_conflict_with_provided_version_fails(self, pacman: Mock):
        """
        If the newest version o package A conflicts with a provided package C (by installed package B),
        but the conflict expression doesn't match (e.g: A conflict expressions (C <= 0.7), C (0.8))
        then B should not be marked as a package to be removed.
        """
        pkg_a = ArchPackage(name="A", version="1.0.0-1", latest_version="1.1.0-1", repository="community")

        pacman.map_provided.side_effect = [{"A": {"A"},  # remote provided
                                            "A=1.1.0": {"A"},
                                            "B": {"B", "C"},
                                            "B=0.8": {"B", "C"},
                                            "C": {"B"}
                                            },
                                           {"A": {"A"},  # provided
                                            "A=1.0.0": {"A"},
                                            "B": {"B", "C"},
                                            "B=0.8": {"B", "C"},
                                            "C": {"B"}
                                            }
                                           ]
        pacman.map_repositories.return_value = {"A": pkg_a.repository, "B": "community"}
        pacman.map_updates_data.return_value = {"A": {'ds': 1,
                                                      's': 1,
                                                      'v': pkg_a.latest_version,
                                                      'c': {"C<=0.7"},
                                                      'p': {"A": {"A"},
                                                            "A=1.1.0": {"A"}},
                                                      'd': set(),
                                                      'r': "community",
                                                      'des': pkg_a.name}}
        pacman.map_installed.return_value = {"A": pkg_a.version, "B": "0.8-1"}
        pacman.get_installed_size.return_value = {"A": 1, "B": 1}
        pacman.map_required_by.return_value = {"A": set()}

        self.deps_analyser.map_missing_deps.return_value = list()

        res = self.summarizer.summarize(pkgs=[pkg_a], root_password=None, arch_config=self.config_)
        for method in ('map_provided', 'map_repositories', 'map_updates_data', 'map_installed',
                       'get_installed_size', 'map_required_by'):
            getattr(pacman, method).assert_called()

        self.assertFalse(res.to_remove)
        self.assertFalse(res.to_install)
        self.assertEqual([UpgradeRequirement(pkg=pkg_a, required_size=1, extra_size=0)], res.to_upgrade)

    @patch(f"{__app_name__}.gems.arch.updates.pacman")
    def test__not_return_to_remove_when_conflict_with_provided_virtual_package_with_version_fails(self, pacman: Mock):
        """
        Scenario:
        - Package V (2.5.0)[update] -> conflicts: X<21.1.1, X-ABI-VIDEODRV_VERSION<25, X-ABI-VIDEODRV_VERSION>=26
        - Package X (21.1.5)[installed] -> provides: X-ABI-VIDEODRV_VERSION=25.2, X

        * X provides X-ABI-VIDEODRV_VERSION with a specific version (25.2)
        """
        pkg_a = ArchPackage(name="V", version="2.5.0-1", latest_version="2.6.0-1", repository="community")
        pacman.map_provided.side_effect = [{"V": {"V"},  # remote provided
                                            "V=2.5.0": {"V"},
                                            "X": {"X"},
                                            "X=21.1.5": {"X"},
                                            "X-ABI-VIDEODRV_VERSION=25.2": {"X"},
                                            "X-ABI-VIDEODRV_VERSION": {"X"}
                                            },
                                           {"V": {"V"},  # provided
                                            "V=2.5.0": {"V"},
                                            "X": {"X"},
                                            "X=21.1.5": {"X"},
                                            "X-ABI-VIDEODRV_VERSION=25.2": {"X"},
                                            "X-ABI-VIDEODRV_VERSION": {"X"}
                                            },
                                           ]
        pacman.map_repositories.return_value = {"V": pkg_a.repository, "X": "community"}
        pacman.map_updates_data.return_value = {"V": {'ds': 1,
                                                      's': 1,
                                                      'v': pkg_a.latest_version,
                                                      'c': {"X<21.1.1",
                                                            "X-ABI-VIDEODRV_VERSION<25",
                                                            "X-ABI-VIDEODRV_VERSION>=26"},
                                                      'p': {"V": {"V"},
                                                            "V=2.5.0": {"A"}},
                                                      'd': set(),
                                                      'r': "community",
                                                      'des': pkg_a.name}}
        pacman.map_installed.return_value = {"V": pkg_a.version, "X": "21.1.5-1"}
        pacman.get_installed_size.return_value = {"V": 1, "X": 1}
        pacman.map_required_by.return_value = {"V": set()}

        self.deps_analyser.map_missing_deps.return_value = list()

        res = self.summarizer.summarize(pkgs=[pkg_a], root_password=None, arch_config=self.config_)
        for method in ('map_provided', 'map_repositories', 'map_updates_data', 'map_installed',
                       'get_installed_size', 'map_required_by'):
            getattr(pacman, method).assert_called()

        self.assertFalse(res.to_remove)
        self.assertFalse(res.to_install)
        self.assertEqual([UpgradeRequirement(pkg=pkg_a, required_size=1, extra_size=0)], res.to_upgrade)

    @patch(f"{__app_name__}.gems.arch.updates.pacman")
    def test__return_installed_virtual_with_defined_version_to_remove_when_conflict_version_matches(self, pacman: Mock):
        """
        Scenario:
        - Package V (2.5.0)[update] -> conflicts: X<21.1.1, X-ABI-VIDEODRV_VERSION<25, X-ABI-VIDEODRV_VERSION>=26
        - Package X (21.1.5)[installed] -> provides: X-ABI-VIDEODRV_VERSION=26, X

        * X provides X-ABI-VIDEODRV_VERSION with a specific version (26)
        """
        pkg_a = ArchPackage(name="V", version="2.5.0-1", latest_version="2.6.0-1", repository="community")

        pacman.map_provided.side_effect = [{"V": {"V"},  # remote provided
                                            "V=2.5.0": {"V"},
                                            "X": {"X"},
                                            "X=21.1.5": {"X"},
                                            "X-ABI-VIDEODRV_VERSION=26": {"X"},
                                            "X-ABI-VIDEODRV_VERSION": {"X"}
                                            },
                                           {"V": {"V"},  # provided
                                            "V=2.5.0": {"V"},
                                            "X": {"X"},
                                            "X=21.1.5": {"X"},
                                            "X-ABI-VIDEODRV_VERSION=26": {"X"},
                                            "X-ABI-VIDEODRV_VERSION": {"X"}
                                            },
                                           {"X": {"X"},   # provided to remove
                                            "X=21.1.5": {"X"},
                                            "X-ABI-VIDEODRV_VERSION=26": {"X"},
                                            "X-ABI-VIDEODRV_VERSION": {"X"}
                                           }
                                           ]
        pacman.map_repositories.return_value = {"V": pkg_a.repository, "X": "community"}
        pacman.map_updates_data.return_value = {"V": {'ds': 1,
                                                      's': 1,
                                                      'v': pkg_a.latest_version,
                                                      'c': {"X<21.1.1",
                                                            "X-ABI-VIDEODRV_VERSION<25",
                                                            "X-ABI-VIDEODRV_VERSION>=26"},
                                                      'p': {"V": {"V"},
                                                            "V=2.5.0": {"A"}},
                                                      'd': set(),
                                                      'r': "community",
                                                      'des': pkg_a.name}}
        pacman.map_installed.return_value = {"V": pkg_a.version, "X": "21.1.5-1"}
        pacman.get_installed_size.return_value = {"V": 1, "X": 1}
        pacman.map_required_by.return_value = {"V": set()}

        self.deps_analyser.map_missing_deps.return_value = list()
        self.deps_analyser.map_all_required_by.return_value = set()

        res = self.summarizer.summarize(pkgs=[pkg_a], root_password=None, arch_config=self.config_)

        for method in ('map_provided', 'map_repositories', 'map_updates_data', 'map_installed',
                       'get_installed_size', 'map_required_by'):
            getattr(pacman, method).assert_called()

        for method in ('map_missing_deps', 'map_all_required_by'):
            getattr(self.deps_analyser, method).assert_called()

        self.assertFalse(res.to_install)
        self.assertEqual([UpgradeRequirement(pkg=pkg_a, required_size=1, extra_size=0)], res.to_upgrade)

        pkg_b = ArchPackage(name="X", installed=True, i18n=self.i18n)
        self.assertEqual([UpgradeRequirement(pkg=pkg_b, reason=" 'V'", extra_size=1)], res.to_remove)

    @patch(f"{__app_name__}.gems.arch.updates.pacman")
    def test__return_installed_virtual_with_defined_version_to_remove_when_conflict_matches__case_2(self, pacman: Mock):
        """
        This test case covers the same scenario as the above, but adds an additional provider for the same
        virtual package with a non-conflicting version

        Scenario:
        - Package V (2.5.0)[update] -> conflicts: X<21.1.1, X-ABI-VIDEODRV_VERSION<25, X-ABI-VIDEODRV_VERSION>=26
        - Package X (21.1.5)[installed] -> provides: X-ABI-VIDEODRV_VERSION=26, X
        - Package Z (2.0.0)[installed] -> provides: X-ABI-VIDEODRV_VERSION=25.2, Z

        * X provides X-ABI-VIDEODRV_VERSION with a specific version (26) [conflict]
        * Z provides X-ABI-VIDEODRV_VERSION with a specific version (25.2) [no conflict]
        """
        pkg_a = ArchPackage(name="V", version="2.5.0-1", latest_version="2.6.0-1", repository="community")

        pacman.map_provided.side_effect = [{"V": {"V"},  # remote provided
                                            "V=2.5.0": {"V"},
                                            "X": {"X"},
                                            "X=21.1.5": {"X"},
                                            "X-ABI-VIDEODRV_VERSION=26": {"X"},
                                            "Z": {"Z"},
                                            "Z=2.0.0": {"X"},
                                            "X-ABI-VIDEODRV_VERSION=25.2": {"Z"},
                                            "X-ABI-VIDEODRV_VERSION": {"X", "Z"},
                                            },
                                           {"V": {"V"},  # provided
                                            "V=2.5.0": {"V"},
                                            "X": {"X"},
                                            "X=21.1.5": {"X"},
                                            "X-ABI-VIDEODRV_VERSION=26": {"X"},
                                            "Z": {"Z"},
                                            "Z=2.0.0": {"X"},
                                            "X-ABI-VIDEODRV_VERSION=25.2": {"Z"},
                                            "X-ABI-VIDEODRV_VERSION": {"X", "Z"},
                                            },
                                           {"X": {"X"},  # provided to remove
                                            "X=21.1.5": {"X"},
                                            "X-ABI-VIDEODRV_VERSION=26": {"X"},
                                            "X-ABI-VIDEODRV_VERSION": {"X", "Z"}
                                            }
                                           ]
        pacman.map_repositories.return_value = {"V": pkg_a.repository, "X": "community"}
        pacman.map_updates_data.return_value = {"V": {'ds': 1,
                                                      's': 1,
                                                      'v': pkg_a.latest_version,
                                                      'c': {"X<21.1.1",
                                                            "X-ABI-VIDEODRV_VERSION<25",
                                                            "X-ABI-VIDEODRV_VERSION>=26"},
                                                      'p': {"V": {"V"},
                                                            "V=2.5.0": {"A"}},
                                                      'd': set(),
                                                      'r': "community",
                                                      'des': pkg_a.name}}
        pacman.map_installed.return_value = {"V": pkg_a.version, "X": "21.1.5-1", "Z": "2.0.0-1"}
        pacman.get_installed_size.return_value = {"V": 1, "X": 1, "Z": 1}
        pacman.map_required_by.return_value = {"V": set()}

        self.deps_analyser.map_missing_deps.return_value = list()
        self.deps_analyser.map_all_required_by.return_value = set()

        res = self.summarizer.summarize(pkgs=[pkg_a], root_password=None, arch_config=self.config_)

        for method in ('map_provided', 'map_repositories', 'map_updates_data', 'map_installed',
                       'get_installed_size', 'map_required_by'):
            getattr(pacman, method).assert_called()

        for method in ('map_missing_deps', 'map_all_required_by'):
            getattr(self.deps_analyser, method).assert_called()

        self.assertFalse(res.to_install)
        self.assertEqual([UpgradeRequirement(pkg=pkg_a, required_size=1, extra_size=0)], res.to_upgrade)

        pkg_b = ArchPackage(name="X", installed=True, i18n=self.i18n)
        self.assertEqual([UpgradeRequirement(pkg=pkg_b, reason=" 'V'", extra_size=1)], res.to_remove)

    @patch(f"{__app_name__}.gems.arch.updates.pacman")
    def test__should_return_installed_package_to_remove_when_conflict_with_provided_matches_version(self, pacman: Mock):
        """
        If the newest version o package A conflicts with a provided package C (by installed package B),
        and the conflict expression matches (e.g: A conflict expressions (C <= 0.7), C (0.7))
        then B should be marked as a package to be removed.
        """
        pkg_a = ArchPackage(name="A", version="1.0.0-1", latest_version="1.1.0-1", repository="community")

        pacman.map_provided.side_effect = [{"A": {"A"},  # remote provided
                                            "A=1.1.0": {"A"},
                                            "B": {"B", "C"},
                                            "B=0.8": {"B", "C"},
                                            "C": {"B"}
                                            },
                                           {"A": {"A"},  # provided
                                            "A=1.0.0": {"A"},
                                            "B": {"B", "C"},
                                            "B=0.7": {"B", "C"},
                                            "C": {"B"}
                                            },
                                           {"B": {"B"},  # provided to remove
                                            "B=0.7": {"B"}
                                            }
                                           ]
        pacman.map_repositories.return_value = {"A": pkg_a.repository, "B": "community"}
        pacman.map_updates_data.return_value = {"A": {'ds': 1,
                                                      's': 1,
                                                      'v': pkg_a.latest_version,
                                                      'c': {"C<=0.7"},
                                                      'p': {"A": {"A"},
                                                            "A=1.1.0": {"A"}},
                                                      'd': set(),
                                                      'r': "community",
                                                      'des': pkg_a.name}}
        pacman.map_installed.return_value = {"A": pkg_a.version, "B": "0.7-1"}
        pacman.get_installed_size.return_value = {"A": 1, "B": 1}
        pacman.map_required_by.return_value = {"A": set()}

        self.deps_analyser.map_missing_deps.return_value = list()
        self.deps_analyser.map_all_required_by.return_value = set()

        res = self.summarizer.summarize(pkgs=[pkg_a], root_password=None, arch_config=self.config_)

        for method in ('map_provided', 'map_repositories', 'map_updates_data', 'map_installed',
                       'get_installed_size', 'map_required_by'):
            getattr(pacman, method).assert_called()

        for method in ('map_missing_deps', 'map_all_required_by'):
            getattr(self.deps_analyser, method).assert_called()

        self.assertFalse(res.to_install)
        self.assertEqual([UpgradeRequirement(pkg=pkg_a, required_size=1, extra_size=0)], res.to_upgrade)

        pkg_b = ArchPackage(name="B", installed=True, i18n=self.i18n)
        self.assertEqual([UpgradeRequirement(pkg=pkg_b, reason=" 'A'", extra_size=1)], res.to_remove)

    @patch(f"{__app_name__}.gems.arch.updates.pacman")
    def test__should_return_installed_package_to_remove_when_conflict_with_provided_matches(self, pacman: Mock):
        """
        If the newest version o package A conflicts with a provided package C (by installed package B),
        then B should be marked as a package to be removed.
        """
        pkg_a = ArchPackage(name="A", version="1.0.0-1", latest_version="1.1.0-1", repository="community")

        pacman.map_provided.side_effect = [{"A": {"A"},  # remote provided
                                            "A=1.1.0": {"A"},
                                            "B": {"B", "C"},
                                            "B=0.8": {"B", "C"},
                                            "C": {"B"}
                                            },
                                           {"A": {"A"},  # provided
                                            "A=1.0.0": {"A"},
                                            "B": {"B", "C"},
                                            "B=0.7": {"B", "C"},
                                            "C": {"B"}
                                            },
                                           {"B": {"B"},  # provided to remove
                                            "B=0.7": {"B"}
                                            }
                                           ]
        pacman.map_repositories.return_value = {"A": pkg_a.repository, "B": "community"}
        pacman.map_updates_data.return_value = {"A": {'ds': 1,
                                                      's': 1,
                                                      'v': pkg_a.latest_version,
                                                      'c': {"C"},
                                                      'p': {"A": {"A"},
                                                            "A=1.1.0": {"A"}},
                                                      'd': set(),
                                                      'r': "community",
                                                      'des': pkg_a.name}}
        pacman.map_installed.return_value = {"A": pkg_a.version, "B": "0.7-1"}
        pacman.get_installed_size.return_value = {"A": 1, "B": 1}
        pacman.map_required_by.return_value = {"A": set()}

        self.deps_analyser.map_missing_deps.return_value = list()
        self.deps_analyser.map_all_required_by.return_value = set()

        res = self.summarizer.summarize(pkgs=[pkg_a], root_password=None, arch_config=self.config_)

        for method in ('map_provided', 'map_repositories', 'map_updates_data', 'map_installed',
                       'get_installed_size', 'map_required_by'):
            getattr(pacman, method).assert_called()

        for method in ('map_missing_deps', 'map_all_required_by'):
            getattr(self.deps_analyser, method).assert_called()

        self.assertFalse(res.to_install)
        self.assertEqual([UpgradeRequirement(pkg=pkg_a, required_size=1, extra_size=0)], res.to_upgrade)

        pkg_b = ArchPackage(name="B", installed=True, i18n=self.i18n)
        self.assertEqual([UpgradeRequirement(pkg=pkg_b, reason=" 'A'", extra_size=1)], res.to_remove)

    @patch(f"{__app_name__}.gems.arch.updates.pacman")
    def test__should_not_return_installed_to_remove_when_conflict_with_provider_is_self_conflict(self, pacman: Mock):
        """
        If the newest version o package A conflicts with a provided package C (by A itself),
        then A should not be marked as a package to be removed.
        """
        pkg_a = ArchPackage(name="A", version="1.0.0-1", latest_version="1.1.0-1", repository="community")

        pacman.map_provided.side_effect = [{"A": {"A"},  # remote provided
                                            "A=1.1.0": {"A"},
                                            "C": {"A"}
                                            },
                                           {"A": {"A"},  # provided
                                            "A=1.0.0": {"A"},
                                            "C": {"A"}
                                            }
                                           ]
        pacman.map_repositories.return_value = {"A": pkg_a.repository}
        pacman.map_updates_data.return_value = {"A": {'ds': 1,
                                                      's': 1,
                                                      'v': pkg_a.latest_version,
                                                      'c': {"C<=1.0.0"},
                                                      'p': {"A": {"A"},
                                                            "A=1.1.0": {"A"},
                                                            "C": {"A"}},
                                                      'd': set(),
                                                      'r': "community",
                                                      'des': pkg_a.name}}
        pacman.map_installed.return_value = {"A": pkg_a.version}
        pacman.get_installed_size.return_value = {"A": 1}
        pacman.map_required_by.return_value = {"A": set()}

        self.deps_analyser.map_missing_deps.return_value = list()

        res = self.summarizer.summarize(pkgs=[pkg_a], root_password=None, arch_config=self.config_)
        for method in ('map_provided', 'map_repositories', 'map_updates_data', 'map_installed',
                       'get_installed_size', 'map_required_by'):
            getattr(pacman, method).assert_called()

        self.assertFalse(res.to_remove)
        self.assertFalse(res.to_install)
        self.assertEqual([UpgradeRequirement(pkg=pkg_a, required_size=1, extra_size=0)], res.to_upgrade)

    @patch(f"{__app_name__}.gems.arch.updates.pacman")
    def test__return_as_cannot_upgrade_when_several_to_upgrade_conflict_with_each_other(self, pacman: Mock):
        """
        Consider package A and B are selected to be upgraded:

        If the newest version o package A conflicts with B, and newest version of B conflicts with A, then
        both should be returned as "cannot upgrade"
        """
        pkg_a = ArchPackage(name="A", version="1.0.0-1", latest_version="1.1.0-1", repository="community")
        pkg_b = ArchPackage(name="B", version="1.0.0-1", latest_version="1.1.0-1", repository="community")

        pacman.map_provided.side_effect = [{"A": {"A"},  # remote provided
                                            "A=1.1.0": {"A"},
                                            "B": {"B"},
                                            "B=1.1.0": {"B"}
                                            },
                                           {"A": {"A"},  # provided
                                            "A=1.0.0": {"A"},
                                            "B": {"B"},
                                            "B=1.0.0": {"B"},
                                            }
                                           ]
        pacman.map_repositories.return_value = {"A": pkg_a.repository, "B": pkg_b.repository}
        pacman.map_updates_data.return_value = {"A": {'ds': 1,
                                                      's': 1,
                                                      'v': pkg_a.latest_version,
                                                      'c': {"B"},
                                                      'p': {"A": {"A"},
                                                            "A=1.1.0": {"A"}},
                                                      'd': set(),
                                                      'r': pkg_a.repository,
                                                      'des': pkg_a.name},
                                                "B": {'ds': 1,
                                                      's': 1,
                                                      'v': pkg_b.latest_version,
                                                      'c': {"A"},
                                                      'p': {"B": {"B"},
                                                            "B=1.1.0": {"B"}},
                                                      'd': set(),
                                                      'r': pkg_a.repository,
                                                      'des': pkg_a.name}
                                                }
        pacman.map_installed.return_value = {"A": pkg_a.version, "B": pkg_b.version}
        self.deps_analyser.map_missing_deps.return_value = list()
        self.deps_analyser.map_all_required_by.return_value = set()

        res = self.summarizer.summarize(pkgs=[pkg_a, pkg_b], root_password=None, arch_config=self.config_)

        for method in ('map_provided', 'map_repositories', 'map_updates_data', 'map_installed'):
            getattr(pacman, method).assert_called()

        for method in ('map_missing_deps',):
            getattr(self.deps_analyser, method).assert_called()

        self.assertFalse(res.to_install)
        self.assertFalse(res.to_upgrade)
        self.assertEqual(2, len(res.cannot_upgrade))
        self.assertIn(UpgradeRequirement(pkg=pkg_a, reason=" 'B'"), res.cannot_upgrade)
        self.assertIn(UpgradeRequirement(pkg=pkg_b, reason=" 'A'"), res.cannot_upgrade)

    @patch(f"{__app_name__}.gems.arch.updates.pacman")
    def test__return_as_cannot_upgrade_when_several_to_upgrade_conflict_with_provided_by_each_other(self, pacman: Mock):
        """
        Consider package A and B are selected to be upgraded:

        If the newest version o package A conflicts with C (provided by B), and newest version of B
        conflicts with D (provided by A), then both should be returned as "cannot upgrade"
        """
        pkg_a = ArchPackage(name="A", version="1.0.0-1", latest_version="1.1.0-1", repository="community")
        pkg_b = ArchPackage(name="B", version="1.0.0-1", latest_version="1.1.0-1", repository="community")

        pacman.map_provided.side_effect = [{"A": {"A"},  # remote provided
                                            "A=1.1.0": {"A"},
                                            "B": {"B"},
                                            "B=1.1.0": {"B"}
                                            },
                                           {"A": {"A"},  # provided
                                            "A=1.0.0": {"A"},
                                            "B": {"B"},
                                            "B=1.0.0": {"B"},
                                            }
                                           ]
        pacman.map_repositories.return_value = {"A": pkg_a.repository, "B": pkg_b.repository}
        pacman.map_updates_data.return_value = {"A": {'ds': 1,
                                                      's': 1,
                                                      'v': pkg_a.latest_version,
                                                      'c': {"C"},
                                                      'p': {"A": {"A"},
                                                            "A=1.1.0": {"A"},
                                                            "D": {"A"}},
                                                      'd': set(),
                                                      'r': pkg_a.repository,
                                                      'des': pkg_a.name},
                                                "B": {'ds': 1,
                                                      's': 1,
                                                      'v': pkg_b.latest_version,
                                                      'c': {"D"},
                                                      'p': {"B": {"B"},
                                                            "B=1.1.0": {"B"},
                                                            "C": {"B"}},
                                                      'd': set(),
                                                      'r': pkg_a.repository,
                                                      'des': pkg_a.name}
                                                }
        pacman.map_installed.return_value = {"A": pkg_a.version, "B": pkg_b.version}
        self.deps_analyser.map_missing_deps.return_value = list()
        self.deps_analyser.map_all_required_by.return_value = set()

        res = self.summarizer.summarize(pkgs=[pkg_a, pkg_b], root_password=None, arch_config=self.config_)

        for method in ('map_provided', 'map_repositories', 'map_updates_data', 'map_installed'):
            getattr(pacman, method).assert_called()

        for method in ('map_missing_deps',):
            getattr(self.deps_analyser, method).assert_called()

        self.assertFalse(res.to_install)
        self.assertFalse(res.to_upgrade)
        self.assertEqual(2, len(res.cannot_upgrade))
        self.assertIn(UpgradeRequirement(pkg=pkg_a, reason=" 'B'"), res.cannot_upgrade)
        self.assertIn(UpgradeRequirement(pkg=pkg_b, reason=" 'A'"), res.cannot_upgrade)

    @patch(f"{__app_name__}.gems.arch.updates.pacman")
    def test__return_as_to_remove_when_a_package_to_upgrade_conflicts_with_another(self, pacman: Mock):
        """
        Consider package A and B are selected to be upgraded:

        If the newest version o package A conflicts with B, then B should be marked as 'to_remove'
        """
        pkg_a = ArchPackage(name="A", version="1.0.0-1", latest_version="1.1.0-1", repository="community")
        pkg_b = ArchPackage(name="B", version="1.0.0-1", latest_version="1.1.0-1", repository="community")

        pacman.map_provided.side_effect = [{"A": {"A"},  # remote provided
                                            "A=1.1.0": {"A"},
                                            "B": {"B"},
                                            "B=1.1.0": {"B"}
                                            },
                                           {"A": {"A"},  # provided
                                            "A=1.0.0": {"A"},
                                            "B": {"B"},
                                            "B=1.0.0": {"B"},
                                            }
                                           ]
        pacman.map_repositories.return_value = {"A": pkg_a.repository, "B": pkg_b.repository}
        pacman.map_updates_data.return_value = {"A": {'ds': 1,
                                                      's': 1,
                                                      'v': pkg_a.latest_version,
                                                      'c': {"B"},
                                                      'p': {"A": {"A"},
                                                            "A=1.1.0": {"A"}},
                                                      'd': set(),
                                                      'r': pkg_a.repository,
                                                      'des': pkg_a.name},
                                                "B": {'ds': 1,
                                                      's': 1,
                                                      'v': pkg_b.latest_version,
                                                      'c': set(),
                                                      'p': {"B": {"B"},
                                                            "B=1.1.0": {"B"}},
                                                      'd': set(),
                                                      'r': pkg_a.repository,
                                                      'des': pkg_a.name}
                                                }
        pacman.map_installed.return_value = {"A": pkg_a.version, "B": pkg_b.version}
        pacman.get_installed_size.return_value = {"A": 1, "B": 1}
        pacman.map_required_by.return_value = {"A": set(), "B": set()}

        self.deps_analyser.map_missing_deps.return_value = list()
        self.deps_analyser.map_all_required_by.return_value = set()

        res = self.summarizer.summarize(pkgs=[pkg_a, pkg_b], root_password=None, arch_config=self.config_)

        for method in ('map_provided', 'map_repositories', 'map_updates_data', 'map_installed',
                       'get_installed_size', 'map_required_by'):
            getattr(pacman, method).assert_called()

        for method in ('map_missing_deps', 'map_all_required_by'):
            getattr(self.deps_analyser, method).assert_called()

        self.assertFalse(res.cannot_upgrade)
        self.assertFalse(res.to_install)
        self.assertEqual([UpgradeRequirement(pkg=pkg_a, required_size=1, extra_size=0)], res.to_upgrade)
        self.assertEqual([UpgradeRequirement(pkg=pkg_b, reason=" 'A'", extra_size=1)], res.to_remove)

    @patch(f"{__app_name__}.gems.arch.updates.pacman")
    def test__return_as_to_remove_when_a_package_to_upgrade_conflicts_with_provided_by_another(self, pacman: Mock):
        """
        Consider package A and B are selected to be upgraded:

        If the newest version o package A conflicts with C (provided by B), then B should be marked as 'to_remove'
        """
        pkg_a = ArchPackage(name="A", version="1.0.0-1", latest_version="1.1.0-1", repository="community")
        pkg_b = ArchPackage(name="B", version="1.0.0-1", latest_version="1.1.0-1", repository="community")

        pacman.map_provided.side_effect = [{"A": {"A"},  # remote provided
                                            "A=1.1.0": {"A"},
                                            "B": {"B"},
                                            "B=1.1.0": {"B"},
                                            "C": {"B"}
                                            },
                                           {"A": {"A"},  # provided
                                            "A=1.0.0": {"A"},
                                            "B": {"B"},
                                            "B=1.0.0": {"B"},
                                            "C": {"B"}
                                            }
                                           ]
        pacman.map_repositories.return_value = {"A": pkg_a.repository, "B": pkg_b.repository}
        pacman.map_updates_data.return_value = {"A": {'ds': 1,
                                                      's': 1,
                                                      'v': pkg_a.latest_version,
                                                      'c': {"C"},
                                                      'p': {"A": {"A"},
                                                            "A=1.1.0": {"A"}},
                                                      'd': set(),
                                                      'r': pkg_a.repository,
                                                      'des': pkg_a.name},
                                                "B": {'ds': 1,
                                                      's': 1,
                                                      'v': pkg_b.latest_version,
                                                      'c': set(),
                                                      'p': {"B": {"B"},
                                                            "B=1.1.0": {"B"},
                                                            "C": {"B"}},
                                                      'd': set(),
                                                      'r': pkg_a.repository,
                                                      'des': pkg_a.name}
                                                }
        pacman.map_installed.return_value = {"A": pkg_a.version, "B": pkg_b.version}
        pacman.get_installed_size.return_value = {"A": 1, "B": 1}
        pacman.map_required_by.return_value = {"A": set(), "B": set()}

        self.deps_analyser.map_missing_deps.return_value = list()
        self.deps_analyser.map_all_required_by.return_value = set()

        res = self.summarizer.summarize(pkgs=[pkg_a, pkg_b], root_password=None, arch_config=self.config_)

        for method in ('map_provided', 'map_repositories', 'map_updates_data', 'map_installed',
                       'get_installed_size', 'map_required_by'):
            getattr(pacman, method).assert_called()

        for method in ('map_missing_deps', 'map_all_required_by'):
            getattr(self.deps_analyser, method).assert_called()

        self.assertFalse(res.cannot_upgrade)
        self.assertFalse(res.to_install)
        self.assertEqual([UpgradeRequirement(pkg=pkg_a, required_size=1, extra_size=0)], res.to_upgrade)
        self.assertEqual([UpgradeRequirement(pkg=pkg_b, reason=" 'A'", extra_size=1)], res.to_remove)

    @patch(f"{__app_name__}.gems.arch.updates.pacman")
    def test__return_as_to_remove_when_a_package_to_upgrade_conflicts_with_provided_version_by_another(self, pacman: Mock):
        """
        Consider package A and B are selected to be upgraded:

        If the newest version o package A conflicts with specific version of C (provided by B),
        then B should be marked as 'to_remove'
        """
        pkg_a = ArchPackage(name="A", version="1.0.0-1", latest_version="1.1.0-1", repository="community")
        pkg_b = ArchPackage(name="B", version="1.0.0-1", latest_version="1.1.0-1", repository="community")

        pacman.map_provided.side_effect = [{"A": {"A"},  # remote provided
                                            "A=1.1.0": {"A"},
                                            "B": {"B"},
                                            "B=1.1.0": {"B"},
                                            "C": {"B"}
                                            },
                                           {"A": {"A"},  # provided
                                            "A=1.0.0": {"A"},
                                            "B": {"B"},
                                            "B=1.0.0": {"B"},
                                            "C": {"B"}
                                            }
                                           ]
        pacman.map_repositories.return_value = {"A": pkg_a.repository, "B": pkg_b.repository}
        pacman.map_updates_data.return_value = {"A": {'ds': 1,
                                                      's': 1,
                                                      'v': pkg_a.latest_version,
                                                      'c': {"C<=1.1.0"},
                                                      'p': {"A": {"A"},
                                                            "A=1.1.0": {"A"}},
                                                      'd': set(),
                                                      'r': pkg_a.repository,
                                                      'des': pkg_a.name},
                                                "B": {'ds': 1,
                                                      's': 1,
                                                      'v': pkg_b.latest_version,
                                                      'c': set(),
                                                      'p': {"B": {"B"},
                                                            "B=1.1.0": {"B"},
                                                            "C": {"B"}},
                                                      'd': set(),
                                                      'r': pkg_a.repository,
                                                      'des': pkg_a.name}
                                                }
        pacman.map_installed.return_value = {"A": pkg_a.version, "B": pkg_b.version}
        pacman.get_installed_size.return_value = {"A": 1, "B": 1}
        pacman.map_required_by.return_value = {"A": set(), "B": set()}

        self.deps_analyser.map_missing_deps.return_value = list()
        self.deps_analyser.map_all_required_by.return_value = set()

        res = self.summarizer.summarize(pkgs=[pkg_a, pkg_b], root_password=None, arch_config=self.config_)

        for method in ('map_provided', 'map_repositories', 'map_updates_data', 'map_installed',
                       'get_installed_size', 'map_required_by'):
            getattr(pacman, method).assert_called()

        for method in ('map_missing_deps', 'map_all_required_by'):
            getattr(self.deps_analyser, method).assert_called()

        self.assertFalse(res.cannot_upgrade)
        self.assertFalse(res.to_install)
        self.assertEqual([UpgradeRequirement(pkg=pkg_a, required_size=1, extra_size=0)], res.to_upgrade)
        self.assertEqual([UpgradeRequirement(pkg=pkg_b, reason=" 'A'", extra_size=1)], res.to_remove)

    @patch(f"{__app_name__}.gems.arch.dependencies.pacman")
    @patch(f"{__app_name__}.gems.arch.updates.pacman")
    def test__return_as_cannot_upgrade_when_several_packages_to_install_conflict_with_each_other(self, *mocks: Mock):
        """
        Consider package A and B are selected to be upgraded:

        If the newest version o package A requires a new package C that conflicts D (new dependency of B),
        and D conflicts with C, then A and B should be marked as 'cannot upgrade'.
        """
        pkg_a = ArchPackage(name="A", version="1.0.0-1", latest_version="1.1.0-1", repository="community")
        pkg_b = ArchPackage(name="B", version="1.0.0-1", latest_version="1.1.0-1", repository="community")

        pacman, pacman_dependencies = mocks[0], mocks[1]
        pacman.map_provided.side_effect = [{"A": {"A"},  # remote provided
                                            "A=1.1.0": {"A"},
                                            "B": {"B"},
                                            "B=1.1.0": {"B"},
                                            "C": {"C"},
                                            "D": {"D"}
                                            },
                                           {"A": {"A"},  # provided
                                            "A=1.0.0": {"A"},
                                            "B": {"B"},
                                            "B=1.0.0": {"B"},
                                            }
                                           ]
        pacman.map_repositories.return_value = {c: "community" for c in ("A", "B", "C", "D")}
        pacman.map_updates_data.side_effect = [{"A": {'ds': 1,
                                                      's': 1,
                                                      'v': pkg_a.latest_version,
                                                      'c': set(),
                                                      'p': {"A": {"A"},
                                                            "A=1.1.0": {"A"}},
                                                      'd': {"C"},
                                                      'r': pkg_a.repository,
                                                      'des': "A"},
                                                "B": {'ds': 1,
                                                      's': 1,
                                                      'v': pkg_b.latest_version,
                                                      'c': set(),
                                                      'p': {"B": {"B"},
                                                            "B=1.1.0": {"B"}},
                                                      'd': {"D"},
                                                      'r': pkg_a.repository,
                                                      'des': "B"}
                                                },
                                               {"C": {'ds': 1,
                                                      's': 1,
                                                      'v': "1.1.0-1",
                                                      'c': {"D"},
                                                      'p': {"C": {"C"},
                                                            "C=1.1.0": {"C"}},
                                                      'd': set(),
                                                      'r': "community",
                                                      'des': "C"},
                                                "D": {'ds': 1,
                                                      's': 1,
                                                      'v': pkg_b.latest_version,
                                                      'c': {"C"},
                                                      'p': {"D": {"D"},
                                                            "D=1.1.0": {"D"}},
                                                      'd': set(),
                                                      'r': "community",
                                                      'des': "D"}
                                                }]
        pacman_dependencies.map_updates_data.return_value = {"C": {'ds': 1,
                                                                   's': 1,
                                                                   'v': "1.1.0-1",
                                                                   'c': {"D"},
                                                                   'p': {"C": {"C"},
                                                                         "C=1.1.0": {"C"}},
                                                                   'd': set(),
                                                                   'r': "community",
                                                                   'des': "C"},
                                                             "D": {'ds': 1,
                                                                   's': 1,
                                                                   'v': pkg_b.latest_version,
                                                                   'c': {"C"},
                                                                   'p': {"D": {"D"},
                                                                         "D=1.1.0": {"D"}},
                                                                   'd': set(),
                                                                   'r': "community",
                                                                   'des': "D"}
                                                             }
        pacman.map_installed.return_value = {"A": pkg_a.version, "B": pkg_b.version}
        pacman.map_required_by.return_value = {c: set() for c in ("A", "B", "C", "D")}
        self.deps_analyser = DependenciesAnalyser(aur_client=self.aur_client, i18n=self.i18n, logger=Mock())
        self.summarizer.deps_analyser = self.deps_analyser

        res = self.summarizer.summarize(pkgs=[pkg_a, pkg_b], root_password=None, arch_config=self.config_)

        for method in ('map_provided', 'map_repositories', 'map_updates_data', 'map_installed'):
            getattr(pacman, method).assert_called()

        pacman_dependencies.map_updates_data.assert_called()

        self.assertFalse(res.to_install)
        self.assertFalse(res.to_upgrade)
        self.assertEqual(2, len(res.cannot_upgrade))
        self.assertIn(UpgradeRequirement(pkg=pkg_a, reason="'C' 'D'"), res.cannot_upgrade)
        self.assertIn(UpgradeRequirement(pkg=pkg_b, reason="'C' 'D'"), res.cannot_upgrade)

    @patch(f"{__app_name__}.gems.arch.dependencies.pacman")
    @patch(f"{__app_name__}.gems.arch.updates.pacman")
    def test__return_as_cannot_upgrade_when_packages_to_install_conflict_with_version_of_each_other(self, *mocks: Mock):
        """
        Consider package A and B are selected to be upgraded:

        If the newest version o package A requires a new package C that conflicts with D's version
        (new dependency of B), and D conflicts with C's version, then A and B should be marked as 'cannot upgrade'.
        """
        pkg_a = ArchPackage(name="A", version="1.0.0-1", latest_version="1.1.0-1", repository="community")
        pkg_b = ArchPackage(name="B", version="1.0.0-1", latest_version="1.1.0-1", repository="community")

        pacman, pacman_dependencies = mocks[0], mocks[1]
        pacman.map_provided.side_effect = [{"A": {"A"},  # remote provided
                                            "A=1.1.0": {"A"},
                                            "B": {"B"},
                                            "B=1.1.0": {"B"},
                                            "C": {"C"},
                                            "D": {"D"}
                                            },
                                           {"A": {"A"},  # provided
                                            "A=1.0.0": {"A"},
                                            "B": {"B"},
                                            "B=1.0.0": {"B"},
                                            }
                                           ]
        pacman.map_repositories.return_value = {c: "community" for c in ("A", "B", "C", "D")}
        pacman.map_updates_data.side_effect = [{"A": {'ds': 1,
                                                      's': 1,
                                                      'v': pkg_a.latest_version,
                                                      'c': set(),
                                                      'p': {"A": {"A"},
                                                            "A=1.1.0": {"A"}},
                                                      'd': {"C"},
                                                      'r': pkg_a.repository,
                                                      'des': "A"},
                                                "B": {'ds': 1,
                                                      's': 1,
                                                      'v': pkg_b.latest_version,
                                                      'c': set(),
                                                      'p': {"B": {"B"},
                                                            "B=1.1.0": {"B"}},
                                                      'd': {"D"},
                                                      'r': pkg_a.repository,
                                                      'des': "B"}
                                                },
                                               {"C": {'ds': 1,
                                                      's': 1,
                                                      'v': "1.1.0-1",
                                                      'c': {"D<=1.1.0"},
                                                      'p': {"C": {"C"},
                                                            "C=1.1.0": {"C"}},
                                                      'd': set(),
                                                      'r': "community",
                                                      'des': "C"},
                                                "D": {'ds': 1,
                                                      's': 1,
                                                      'v': pkg_b.latest_version,
                                                      'c': {"C<=1.1.0"},
                                                      'p': {"D": {"D"},
                                                            "D=1.1.0": {"D"}},
                                                      'd': set(),
                                                      'r': "community",
                                                      'des': "D"}
                                                }]
        pacman_dependencies.map_updates_data.return_value = {"C": {'ds': 1,
                                                                   's': 1,
                                                                   'v': "1.1.0-1",
                                                                   'c': {"D<=1.1.0"},
                                                                   'p': {"C": {"C"},
                                                                         "C=1.1.0": {"C"}},
                                                                   'd': set(),
                                                                   'r': "community",
                                                                   'des': "C"},
                                                             "D": {'ds': 1,
                                                                   's': 1,
                                                                   'v': pkg_b.latest_version,
                                                                   'c': {"C<=1.1.0"},
                                                                   'p': {"D": {"D"},
                                                                         "D=1.1.0": {"D"}},
                                                                   'd': set(),
                                                                   'r': "community",
                                                                   'des': "D"}
                                                             }
        pacman.map_installed.return_value = {"A": pkg_a.version, "B": pkg_b.version}
        pacman.map_required_by.return_value = {c: set() for c in ("A", "B", "C", "D")}
        self.deps_analyser = DependenciesAnalyser(aur_client=self.aur_client, i18n=self.i18n, logger=Mock())
        self.summarizer.deps_analyser = self.deps_analyser

        res = self.summarizer.summarize(pkgs=[pkg_a, pkg_b], root_password=None, arch_config=self.config_)

        for method in ('map_provided', 'map_repositories', 'map_updates_data', 'map_installed'):
            getattr(pacman, method).assert_called()

        pacman_dependencies.map_updates_data.assert_called()

        self.assertFalse(res.to_install)
        self.assertFalse(res.to_upgrade)
        self.assertEqual(2, len(res.cannot_upgrade))
        self.assertIn(UpgradeRequirement(pkg=pkg_a, reason="'C' 'D'"), res.cannot_upgrade)
        self.assertIn(UpgradeRequirement(pkg=pkg_b, reason="'C' 'D'"), res.cannot_upgrade)

    @patch(f"{__app_name__}.gems.arch.dependencies.pacman")
    @patch(f"{__app_name__}.gems.arch.updates.pacman")
    def test__return_as_to_remove_when_to_update_conflicts_with_to_install(self, *mocks: Mock):
        """
        Consider package A and B are selected to be upgraded:

        If the newest version o package A requires a new package C that conflicts with B,
        then B should be marked as "to_remove"
        """
        pkg_a = ArchPackage(name="A", version="1.0.0-1", latest_version="1.1.0-1", repository="community")
        pkg_b = ArchPackage(name="B", version="1.0.0-1", latest_version="1.1.0-1", repository="community")

        pacman, pacman_dependencies = mocks[0], mocks[1]
        pacman.map_provided.side_effect = [{"A": {"A"},  # remote provided
                                            "A=1.1.0": {"A"},
                                            "B": {"B"},
                                            "B=1.1.0": {"B"},
                                            "C": {"C"}
                                            },
                                           {"A": {"A"},  # provided
                                            "A=1.0.0": {"A"},
                                            "B": {"B"},
                                            "B=1.0.0": {"B"},
                                            }
                                           ]
        pacman.map_repositories.return_value = {c: "community" for c in ("A", "B", "C")}
        pacman.map_updates_data.side_effect = [{"A": {'ds': 1,
                                                      's': 1,
                                                      'v': pkg_a.latest_version,
                                                      'c': set(),
                                                      'p': {"A": {"A"},
                                                            "A=1.1.0": {"A"}},
                                                      'd': {"C"},
                                                      'r': pkg_a.repository,
                                                      'des': "A"},
                                                "B": {'ds': 1,
                                                      's': 1,
                                                      'v': pkg_b.latest_version,
                                                      'c': set(),
                                                      'p': {"B": {"B"},
                                                            "B=1.1.0": {"B"}},
                                                      'd': set(),
                                                      'r': pkg_a.repository,
                                                      'des': "B"}
                                                },
                                               {"C": {'ds': 1,
                                                      's': 1,
                                                      'v': "1.1.0-1",
                                                      'c': {"B"},
                                                      'p': {"C": {"C"},
                                                            "C=1.1.0": {"C"}},
                                                      'd': set(),
                                                      'r': "community",
                                                      'des': "C"}
                                                }
                                               ]
        pacman_dependencies.map_updates_data.return_value = {"C": {'ds': 1,
                                                                   's': 1,
                                                                   'v': "1.1.0-1",
                                                                   'c': {"B"},
                                                                   'p': {"C": {"C"},
                                                                         "C=1.1.0": {"C"}},
                                                                   'd': set(),
                                                                   'r': "community",
                                                                   'des': "C"}
                                                             }
        pacman.map_installed.return_value = {"A": pkg_a.version, "B": pkg_b.version}
        pacman.map_required_by.return_value = {c: set() for c in ("A", "B", "C")}
        pacman.get_installed_size.return_value = {"A": 1, "B": 1, "C": 1}
        self.deps_analyser = DependenciesAnalyser(aur_client=self.aur_client, i18n=self.i18n, logger=Mock())
        self.summarizer.deps_analyser = self.deps_analyser

        res = self.summarizer.summarize(pkgs=[pkg_a, pkg_b], root_password=None, arch_config=self.config_)

        for method in ('map_provided', 'map_repositories', 'map_updates_data', 'map_installed', 'get_installed_size',
                       'map_required_by'):
            getattr(pacman, method).assert_called()

        pacman_dependencies.map_updates_data.assert_called()

        self.assertFalse(res.cannot_upgrade)

        pkg_c = ArchPackage(name="C", version="1.0.0-1", latest_version="1.1.0-1", repository="community")
        self.assertEqual([UpgradeRequirement(pkg=pkg_c, required_size=1, extra_size=1, reason=": A")], res.to_install)
        self.assertEqual([UpgradeRequirement(pkg=pkg_a, required_size=1, extra_size=0)], res.to_upgrade)
        self.assertEqual([UpgradeRequirement(pkg=pkg_b, extra_size=1, reason=" 'C'")], res.to_remove)

    @patch(f"{__app_name__}.gems.arch.dependencies.pacman")
    @patch(f"{__app_name__}.gems.arch.updates.pacman")
    def test__return_as_to_remove_when_to_update_conflicts_with_specific_version_of_install(self, *mocks: Mock):
        """
        Consider package A and B are selected to be upgraded:

        If the newest version o package A requires a new package C that conflicts with latest version of B,
        then B should be marked as 'to_remove'.
        """
        pkg_a = ArchPackage(name="A", version="1.0.0-1", latest_version="1.1.0-1", repository="community")
        pkg_b = ArchPackage(name="B", version="1.0.0-1", latest_version="1.1.0-1", repository="community")

        pacman, pacman_dependencies = mocks[0], mocks[1]
        pacman.map_provided.side_effect = [{"A": {"A"},  # remote provided
                                            "A=1.1.0": {"A"},
                                            "B": {"B"},
                                            "B=1.1.0": {"B"},
                                            "C": {"C"}
                                            },
                                           {"A": {"A"},  # provided
                                            "A=1.0.0": {"A"},
                                            "B": {"B"},
                                            "B=1.0.0": {"B"},
                                            }
                                           ]
        pacman.map_repositories.return_value = {c: "community" for c in ("A", "B", "C")}
        pacman.map_updates_data.side_effect = [{"A": {'ds': 1,
                                                      's': 1,
                                                      'v': pkg_a.latest_version,
                                                      'c': set(),
                                                      'p': {"A": {"A"},
                                                            "A=1.1.0": {"A"}},
                                                      'd': {"C"},
                                                      'r': pkg_a.repository,
                                                      'des': "A"},
                                                "B": {'ds': 1,
                                                      's': 1,
                                                      'v': pkg_b.latest_version,
                                                      'c': set(),
                                                      'p': {"B": {"B"},
                                                            "B=1.1.0": {"B"}},
                                                      'd': set(),
                                                      'r': pkg_a.repository,
                                                      'des': "B"}
                                                },
                                               {"C": {'ds': 1,
                                                      's': 1,
                                                      'v': "1.1.0-1",
                                                      'c': {"B<=1.1.0"},
                                                      'p': {"C": {"C"},
                                                            "C=1.1.0": {"C"}},
                                                      'd': set(),
                                                      'r': "community",
                                                      'des': "C"}
                                                }
                                               ]
        pacman_dependencies.map_updates_data.return_value = {"C": {'ds': 1,
                                                                   's': 1,
                                                                   'v': "1.1.0-1",
                                                                   'c': {"B<=1.1.0"},
                                                                   'p': {"C": {"C"},
                                                                         "C=1.1.0": {"C"}},
                                                                   'd': set(),
                                                                   'r': "community",
                                                                   'des': "C"}
                                                             }
        pacman.map_installed.return_value = {"A": pkg_a.version, "B": pkg_b.version}
        pacman.map_required_by.return_value = {c: set() for c in ("A", "B", "C")}
        pacman.get_installed_size.return_value = {"A": 1, "B": 1, "C": 1}
        self.deps_analyser = DependenciesAnalyser(aur_client=self.aur_client, i18n=self.i18n, logger=Mock())
        self.summarizer.deps_analyser = self.deps_analyser

        res = self.summarizer.summarize(pkgs=[pkg_a, pkg_b], root_password=None, arch_config=self.config_)

        for method in ('map_provided', 'map_repositories', 'map_updates_data', 'map_installed', 'get_installed_size',
                       'map_required_by'):
            getattr(pacman, method).assert_called()

        pacman_dependencies.map_updates_data.assert_called()

        self.assertFalse(res.cannot_upgrade)

        pkg_c = ArchPackage(name="C", version="1.0.0-1", latest_version="1.1.0-1", repository="community")
        self.assertEqual([UpgradeRequirement(pkg=pkg_c, required_size=1, extra_size=1, reason=": A")], res.to_install)
        self.assertEqual([UpgradeRequirement(pkg=pkg_a, required_size=1, extra_size=0)], res.to_upgrade)
        self.assertEqual([UpgradeRequirement(pkg=pkg_b, extra_size=1, reason=" 'C'")], res.to_remove)

    @patch(f"{__app_name__}.gems.arch.dependencies.pacman")
    @patch(f"{__app_name__}.gems.arch.updates.pacman")
    def test__return_as_to_remove_when_to_update_conflicts_with_to_install_and_it_has_deps(self, *mocks: Mock):
        """
        Consider package A and B are selected to be upgraded:

        If the newest version o package A requires a new package C that conflicts with B,
        then B should be marked as "to_remove" and its dependents as well (package 'D').
        """
        pkg_a = ArchPackage(name="A", version="1.0.0-1", latest_version="1.1.0-1", repository="community")
        pkg_b = ArchPackage(name="B", version="1.0.0-1", latest_version="1.1.0-1", repository="community")

        pacman, pacman_dependencies = mocks[0], mocks[1]
        pacman.map_provided.side_effect = [{"A": {"A"},  # remote provided
                                            "A=1.1.0": {"A"},
                                            "B": {"B"},
                                            "B=1.1.0": {"B"},
                                            "C": {"C"}
                                            },
                                           {"A": {"A"},  # provided
                                            "A=1.0.0": {"A"},
                                            "B": {"B"},
                                            "B=1.0.0": {"B"},
                                            "D": {"D"},
                                            "D=0.7.0": {"D"}
                                            },
                                           {"B": {"B"},  # provided to remove
                                            "B=1.0.0": {"B"},
                                            "D": {"D"},
                                            "D=0.7.0": {"D"}
                                            }
                                           ]
        pacman.map_repositories.return_value = {c: "community" for c in ("A", "B", "C", "D")}
        pacman.map_updates_data.side_effect = [{"A": {'ds': 1,
                                                      's': 1,
                                                      'v': pkg_a.latest_version,
                                                      'c': set(),
                                                      'p': {"A": {"A"},
                                                            "A=1.1.0": {"A"}},
                                                      'd': {"C"},
                                                      'r': pkg_a.repository,
                                                      'des': "A"},
                                                "B": {'ds': 1,
                                                      's': 1,
                                                      'v': pkg_b.latest_version,
                                                      'c': set(),
                                                      'p': {"B": {"B"},
                                                            "B=1.1.0": {"B"}},
                                                      'd': set(),
                                                      'r': pkg_a.repository,
                                                      'des': "B"}
                                                },
                                               {"C": {'ds': 1,
                                                      's': 1,
                                                      'v': "1.1.0-1",
                                                      'c': {"B"},
                                                      'p': {"C": {"C"},
                                                            "C=1.1.0": {"C"}},
                                                      'd': set(),
                                                      'r': "community",
                                                      'des': "C"}
                                                }
                                               ]
        pacman_dependencies.map_updates_data.return_value = {"C": {'ds': 1,
                                                                   's': 1,
                                                                   'v': "1.1.0-1",
                                                                   'c': {"B"},
                                                                   'p': {"C": {"C"},
                                                                         "C=1.1.0": {"C"}},
                                                                   'd': set(),
                                                                   'r': "community",
                                                                   'des': "C"}
                                                             }
        pacman_dependencies.map_required_by.return_value = {"B": {"D"}}

        pacman.map_installed.return_value = {"A": pkg_a.version, "B": pkg_b.version, "D": "0.7.0"}
        pacman.map_required_by.return_value = {**{c: set() for c in ("A", "C", "D")}, "B": {"D"}}
        pacman.map_required_dependencies.return_value = {"D": {"B"}}
        pacman.get_installed_size.return_value = {"A": 1, "B": 1, "C": 1, "D": 1}

        self.deps_analyser = DependenciesAnalyser(aur_client=self.aur_client, i18n=self.i18n, logger=Mock())
        self.summarizer.deps_analyser = self.deps_analyser

        res = self.summarizer.summarize(pkgs=[pkg_a, pkg_b], root_password=None, arch_config=self.config_)

        for method in ('map_provided', 'map_repositories', 'map_updates_data', 'map_installed', 'get_installed_size',
                       'map_required_by', 'map_required_dependencies'):
            getattr(pacman, method).assert_called()

        for method in ('map_updates_data', 'map_required_by'):
            getattr(pacman_dependencies, method).assert_called()

        self.assertFalse(res.cannot_upgrade)

        self.assertEqual([UpgradeRequirement(pkg=pkg_a, required_size=1, extra_size=0)], res.to_upgrade)

        pkg_c = ArchPackage(name="C", version="1.0.0-1", latest_version="1.1.0-1", repository="community")
        self.assertEqual([UpgradeRequirement(pkg=pkg_c, required_size=1, extra_size=1, reason=": A")], res.to_install)

        self.assertEqual(2, len(res.to_remove))
        self.assertIn(UpgradeRequirement(pkg=pkg_b, extra_size=1, reason=" 'C'"), res.to_remove)

        pkg_d = ArchPackage(name="D", repository="community", version="0.7.0-1", latest_version="0.7.0-1")
        self.assertIn(UpgradeRequirement(pkg=pkg_d, extra_size=1, reason=" 'B=1.0.0'"), res.to_remove)



from collections import defaultdict
from typing import Dict, List, Generator, Tuple, Optional, Union

from bauh.view.qt.commons import PackageFilters
from bauh.view.qt.view_model import PackageView


def new_character_idx() -> Dict[str, List[PackageView]]:
    return defaultdict(list)


def new_category_idx() -> Dict[str, Dict[str, List[PackageView]]]:
    return defaultdict(new_character_idx)


def new_type_index() -> Dict[str, Dict[str, Dict[str, List[PackageView]]]]:
    return defaultdict(new_category_idx)


def new_verified_index() -> Dict[int, Dict[str, Dict[str, Dict[str, List[PackageView]]]]]:
    return defaultdict(new_type_index)


def new_update_index() -> Dict[int, Dict[int, Dict[str, Dict[str, Dict[str, List[PackageView]]]]]]:
    return defaultdict(new_verified_index)


def new_app_index() -> Dict[int, Dict[int, Dict[int, Dict[str, Dict[str, Dict[str, List[PackageView]]]]]]]:
    return defaultdict(new_update_index)


def new_package_index() -> Dict[int, Dict[int, Dict[int, Dict[int, Dict[str, Dict[str, Dict[str, List[PackageView]]]]]]]]:
    return defaultdict(new_app_index)


def add_to_index(pkgv: PackageView, index: dict) -> None:
    # root keys: (1) installed | (0) not installed
    root_idx = index[1 if pkgv.model.installed else 0]

    # app keys: (1) app | (0) not app
    app_lvl = root_idx[1 if pkgv.model.is_application() else 0]

    # update keys: (1) update | (0) no update
    if pkgv.model.installed and pkgv.model.update and not pkgv.model.is_update_ignored():
        update_lvl = app_lvl[1]
    else:
        update_lvl = app_lvl[0]

    # verified keys: (1) verified | (0) unverified
    verified_lvl = update_lvl[1 if pkgv.model.is_trustable() else 0]

    norm_name = pkgv.name.strip().lower()
    starts_with_chars = tuple(norm_name[0:i] for i in range(1, len(norm_name) + 1))

    for cat in ("any", *(pkgv.model.categories if pkgv.model.categories else tuple())):
        category = cat.lower().strip()

        # any type > specific category > any character (None)
        verified_lvl["any"][category][None].append(pkgv)

        # any type > specific category > characters (start
        for chars in starts_with_chars:
            verified_lvl["any"][category][chars].append(pkgv)

        type_lvl = verified_lvl[pkgv.model.get_type()]

        # specific type > specific category > any character (None)
        type_lvl[category][None].append(pkgv)

        # specific type > any category > first character
        for chars in starts_with_chars:
            type_lvl[category][chars].append(pkgv)


def generate_queries(filters: PackageFilters) -> Generator[Tuple[Optional[Union[int, str]], ...], None, None]:
    chars_query = None

    if filters.name:
        chars_query = filters.name.lower()

    installed_queries = (1,) if filters.only_installed else (1, 0)
    apps_queries = (1,) if filters.only_apps else (1, 0)
    updates_queries = (1,) if filters.only_updates else (1, 0)
    verified_queries = (1,) if filters.only_verified else (1, 0)

    for installed in installed_queries:
        for app in apps_queries:
            for update in updates_queries:
                for verified in verified_queries:
                    yield installed, app, update, verified, filters.type, filters.category, chars_query


def query_packages(index: dict, filters: PackageFilters) -> Generator[PackageView, None, None]:
    yield_count = 0
    yield_limit = filters.display_limit if filters.display_limit and filters.display_limit > 0 else -1

    queries = tuple(generate_queries(filters))

    yielded_pkgs = defaultdict(set)

    for query in queries:
        packages = index[query[0]][query[1]][query[2]][query[3]][query[4]][query[5]][query[6]]

        for pkgv in packages:
            yield pkgv
            yield_count += 1
            yielded_pkgs[pkgv.model.get_type()].add(pkgv.model.id)

            # checking if the package display limit has been reached
            if 0 < yield_limit <= yield_count:
                break

    # if there is a limit and the number of yielded packages is not reached, performs also a "contains" query
    # checking if the queries target "any character" (none), if so, there is no need to perform the "contains" query
    any_char_query = next((True for q in queries if q[-1] is None), False)

    if not any_char_query and 0 < yield_limit > yield_count:
        for query in queries:
            # checking if the package display limit has been reached
            if 0 < yield_limit <= yield_count:
                break

            packages = index[query[0]][query[1]][query[2]][query[3]][query[4]][query[5]][None]

            for pkgv in packages:
                # checking if the package has already been yielded
                yield_type_idx = yielded_pkgs.get(pkgv.model.get_type())
                if yield_type_idx and pkgv.model.id in yield_type_idx:
                    continue

                # checking if the package name contains the chars query
                if query[5] in pkgv.model.name:
                    yield pkgv
                    yield_count += 1
                    yielded_pkgs[pkgv.model.get_type()].add(pkgv.model.id)

                # checking if the package display limit has been reached
                if 0 < yield_limit <= yield_count:
                    break



import logging
import traceback
from io import BytesIO
from threading import Thread
from typing import List, Dict

from PyQt5.QtCore import Qt
from PyQt5.QtGui import QIcon, QPixmap, QCursor
from PyQt5.QtWidgets import QDialog, QLabel, QPushButton, QVBoxLayout, QProgressBar, QApplication, QWidget, \
    QSizePolicy, QHBoxLayout

from bauh.api.abstract.cache import MemoryCache
from bauh.api.http import HttpClient
from bauh.view.qt import qt_utils
from bauh.view.qt.components import new_spacer
from bauh.view.qt.thread import AnimateProgress
from bauh.view.qt.view_model import PackageView
from bauh.view.util.translation import I18n


class ScreenshotsDialog(QDialog):

    def __init__(self, pkg: PackageView, http_client: HttpClient, icon_cache: MemoryCache, i18n: I18n, screenshots: List[QPixmap], logger: logging.Logger):
        super(ScreenshotsDialog, self).__init__()
        self.setWindowTitle(str(pkg))
        self.screenshots = screenshots
        self.logger = logger
        self.loaded_imgs = []
        self.download_threads = []
        self.i18n = i18n
        self.http_client = http_client
        self.progress_bar = QProgressBar()
        self.progress_bar.setObjectName('progress_screenshots')
        self.progress_bar.setCursor(QCursor(Qt.WaitCursor))
        self.progress_bar.setMaximumHeight(10 if QApplication.instance().style().objectName().lower() == 'windows' else 6)
        self.progress_bar.setTextVisible(False)
        self.thread_progress = AnimateProgress()
        self.thread_progress.signal_change.connect(self._update_progress)
        self.thread_progress.start()

        # THERE ARE CRASHES WITH SOME RARE ICONS ( like insomnia ). IT CAN BE A QT BUG. IN THE MEANTIME, ONLY THE TYPE ICON WILL BE RENDERED
        #
        # icon_data = icon_cache.get(pkg.model.icon_url)
        #
        # if icon_data and icon_data.get('icon'):
        #     self.setWindowIcon(icon_data.get('icon'))
        # else:
        #     self.setWindowIcon(QIcon(pkg.model.get_type_icon_path()))
        self.setWindowIcon(QIcon(pkg.model.get_type_icon_path()))
        self.setLayout(QVBoxLayout())

        self.bt_close = QPushButton(self.i18n['screenshots.bt_close'])
        self.bt_close.setObjectName('close')
        self.bt_close.clicked.connect(self.close)
        self.bt_close.setCursor(QCursor(Qt.PointingHandCursor))

        self.upper_buttons = QWidget()
        self.upper_buttons.setObjectName('upper_buttons')
        self.upper_buttons.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Fixed)
        self.upper_buttons.setContentsMargins(0, 0, 0, 0)
        self.upper_buttons.setLayout(QHBoxLayout())
        self.upper_buttons.layout().setAlignment(Qt.AlignRight)
        self.upper_buttons.layout().addWidget(self.bt_close)
        self.layout().addWidget(self.upper_buttons)

        self.layout().addWidget(new_spacer())

        self.img = QLabel()
        self.img.setObjectName('image')
        self.layout().addWidget(self.img)
        self.layout().addWidget(new_spacer())

        self.container_buttons = QWidget()
        self.container_buttons.setObjectName('buttons_container')
        self.container_buttons.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Fixed)
        self.container_buttons.setContentsMargins(0, 0, 0, 0)
        self.container_buttons.setLayout(QHBoxLayout())

        self.bt_back = QPushButton(' < ' + self.i18n['screenshots.bt_back.label'].capitalize())
        self.bt_back.setObjectName('back')
        self.bt_back.setProperty('control', 'true')
        self.bt_back.setCursor(QCursor(Qt.PointingHandCursor))
        self.bt_back.clicked.connect(self.back)
        self.container_buttons.layout().addWidget(self.bt_back)
        self.container_buttons.layout().addWidget(new_spacer())

        self.img_label = QLabel()
        self.img_label.setObjectName("image_label")
        self.container_buttons.layout().addWidget(self.img_label)

        self.container_buttons.layout().addWidget(self.progress_bar)
        self.container_buttons.layout().addWidget(new_spacer())

        self.bt_next = QPushButton(self.i18n['screenshots.bt_next.label'].capitalize() + ' > ')
        self.bt_next.setObjectName('next')
        self.bt_next.setProperty('control', 'true')
        self.bt_next.setCursor(QCursor(Qt.PointingHandCursor))
        self.bt_next.clicked.connect(self.next)
        self.container_buttons.layout().addWidget(self.bt_next)
        self.download_progress: Dict[int, float] = dict()

        self.layout().addWidget(self.container_buttons)

        self.img_idx = 0
        self.max_img_width = 800
        self.max_img_height = 600

        for idx, s in enumerate(self.screenshots):
            t = Thread(target=self._download_img, args=(idx, s), daemon=True)
            t.start()

        self.resize(self.max_img_width + 5, self.max_img_height + 5)
        self._load_img(self.img_idx)
        qt_utils.centralize(self)

    def _update_progress(self, val: int):
        self.progress_bar.setValue(val)

    def _load_img(self, img_idx: int):
        if img_idx != self.img_idx:
            return

        if len(self.loaded_imgs) > self.img_idx:
            img = self.loaded_imgs[self.img_idx]

            if isinstance(img, QPixmap):
                self.img_label.setText(f'{self.img_idx + 1}/{len(self.screenshots)}')
                self.img.setText('')
                self.img.setPixmap(img)
            else:
                self.img.setText(img)
                self.img.setPixmap(QPixmap())

            self.img.unsetCursor()
            self.thread_progress.stop = True
            self.progress_bar.setVisible(False)
        else:
            self.img.setPixmap(QPixmap())
            self.img.setCursor(QCursor(Qt.WaitCursor))

            progress = self.download_progress.get(self.img_idx, 0)
            self.img.setText(f"{self.i18n['screenshots.image.loading']} "
                             f"{self.img_idx + 1}/{len(self.screenshots)} ({progress:.2f}%)")
            self.progress_bar.setVisible(True)
            self.thread_progress.start()

        if len(self.screenshots) == 1:
            self.bt_back.setVisible(False)
            self.bt_next.setVisible(False)
        else:
            self.bt_back.setEnabled(self.img_idx != 0)
            self.bt_next.setEnabled(self.img_idx != len(self.screenshots) - 1)

    def _handle_download_exception(self, idx: int, url: str):
        self.logger.error(f"Unexpected exception while downloading screenshot from '{url}'")
        traceback.print_exc()
        self.loaded_imgs.append(self.i18n["screenshots.download.no_response"])
        self._load_img(idx)

    def _download_img(self, idx: int, url: str):
        self.logger.info(f"Downloading image [{idx}] from {url}")

        try:
            res = self.http_client.get(url=url, stream=True)
        except Exception:
            self._handle_download_exception(idx, url)
            return

        if not res:
            self.logger.info(f"Could not retrieve image [{idx}] from '{url}'")
            self.loaded_imgs.append(self.i18n["screenshots.download.no_response"])
            self._load_img(idx)
            return

        try:
            content_length = int(res.headers.get("content-length", 0))
        except Exception:
            content_length = 0
            self.logger.warning(f"Could not retrieve the content-length for file '{url}'")

        if content_length <= 0:
            self.logger.warning(f"Image [{idx}] has no content ({url})")
            self.loaded_imgs.append(self.i18n['screenshots.download.no_content'])
            self._load_img(idx)
        else:
            byte_stream = BytesIO()

            total_downloaded = 0
            try:
                for data in res.iter_content(chunk_size=1024):
                    byte_stream.write(data)
                    total_downloaded += len(data)
                    self.download_progress[idx] = (total_downloaded / content_length) * 100
                    self._load_img(idx)
            except Exception:
                self._handle_download_exception(idx, url)
                return

            self.logger.info(f"Image [{idx}] successfully downloaded ({url})")
            pixmap = QPixmap()
            pixmap.loadFromData(byte_stream.getvalue())

            if pixmap.size().height() > self.max_img_height or pixmap.size().width() > self.max_img_width:
                pixmap = pixmap.scaled(self.max_img_width, self.max_img_height, Qt.KeepAspectRatio,
                                       Qt.SmoothTransformation)

            self.loaded_imgs.append(pixmap)
            self._load_img(idx)

    def back(self):
        self.img_idx -= 1
        self._load_img(self.img_idx)

    def next(self):
        self.img_idx += 1
        self._load_img(self.img_idx)



import shutil
import subprocess
from collections.abc import Iterable
from subprocess import Popen
from typing import Optional

from PyQt5.QtCore import Qt
from PyQt5.QtGui import QIcon, QCursor
from PyQt5.QtWidgets import QDialog, QVBoxLayout, QGroupBox, \
    QLineEdit, QLabel, QGridLayout, QPushButton, QPlainTextEdit, QScrollArea, QFrame, QWidget, QSizePolicy, \
    QHBoxLayout

from bauh.api.abstract.cache import MemoryCache
from bauh.commons.regex import RE_URL
from bauh.view.qt.components import new_spacer
from bauh.view.qt.qt_utils import get_current_screen_geometry
from bauh.view.util.translation import I18n

IGNORED_ATTRS = {'name', '__app__'}


class InfoDialog(QDialog):

    def __init__(self, pkg_info: dict, icon_cache: MemoryCache, i18n: I18n, can_open_url: bool):
        super(InfoDialog, self).__init__()
        self.setWindowTitle(str(pkg_info['__app__']))
        self.i18n = i18n
        self._can_open_url = can_open_url
        layout = QVBoxLayout()
        self.setLayout(layout)

        scroll = QScrollArea(self)
        scroll.setFrameShape(QFrame.NoFrame)
        scroll.setWidgetResizable(True)
        comps_container = QWidget()
        comps_container.setObjectName('root_container')
        comps_container.setLayout(QVBoxLayout())
        scroll.setWidget(comps_container)

        # shows complete field string
        self.text_field = QPlainTextEdit()
        self.text_field.setObjectName('full_field')
        self.text_field.setReadOnly(True)
        comps_container.layout().addWidget(self.text_field)
        self.text_field.hide()

        self.gbox_info = QGroupBox()
        self.gbox_info.setObjectName('fields')
        self.gbox_info.setLayout(QGridLayout())

        comps_container.layout().addWidget(self.gbox_info)

        # THERE ARE CRASHES WITH SOME RARE ICONS ( like insomnia ). IT CAN BE A QT BUG. IN THE MEANTIME, ONLY THE TYPE ICON WILL BE RENDERED
        #
        # icon_data = icon_cache.get(app['__app__'].model.icon_url)
        #
        # if icon_data and icon_data.get('icon'):
        #     self.setWindowIcon(icon_data.get('icon'))
        self.setWindowIcon(QIcon(pkg_info['__app__'].model.get_type_icon_path()))

        for idx, attr in enumerate(sorted(pkg_info.keys())):
            if attr not in IGNORED_ATTRS and pkg_info[attr] is not None:
                i18n_key = pkg_info['__app__'].model.gem_name + '.info.' + attr.lower()
                val = pkg_info[attr]

                if not isinstance(val, str) and isinstance(pkg_info[attr], Iterable):
                    val = ' '.join([str(e).strip() for e in pkg_info[attr] if e])
                    show_val = '\n'.join(['* ' + str(e).strip() for e in pkg_info[attr] if e])
                else:
                    val = str(pkg_info[attr]).strip()
                    show_val = val

                i18n_val = i18n.get(f"{i18n_key}.{val.lower()}")

                if i18n_val:
                    val = i18n_val
                    show_val = val

                text = QLineEdit()
                text.setObjectName('field_value')
                text.setToolTip(show_val)
                text.setText(val)
                text.setCursorPosition(0)
                text.setReadOnly(True)

                label = QLabel(i18n.get(i18n_key, i18n.get(attr.lower(), attr)).capitalize())
                label.setObjectName('field_name')

                self.gbox_info.layout().addWidget(label, idx, 0)
                self.gbox_info.layout().addWidget(text, idx, 1)

                self._gen_show_button(idx=idx, val=show_val)

        layout.addWidget(scroll)

        lower_container = QWidget()
        lower_container.setObjectName('lower_container')
        lower_container.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Fixed)
        lower_container.setLayout(QHBoxLayout())

        self.bt_back = QPushButton('< {}'.format(self.i18n['back'].capitalize()))
        self.bt_back.setObjectName('back')
        self.bt_back.setVisible(False)
        self.bt_back.setCursor(QCursor(Qt.PointingHandCursor))
        self.bt_back.clicked.connect(self.back_to_info)

        lower_container.layout().addWidget(self.bt_back)
        lower_container.layout().addWidget(new_spacer())

        self.bt_close = QPushButton(self.i18n['close'].capitalize())
        self.bt_close.setObjectName('close')
        self.bt_close.setCursor(QCursor(Qt.PointingHandCursor))
        self.bt_close.clicked.connect(self.close)

        lower_container.layout().addWidget(self.bt_close)
        layout.addWidget(lower_container)
        self.setMinimumWidth(int(self.gbox_info.sizeHint().width() * 1.2))

        screen_height = get_current_screen_geometry().height()
        self.setMaximumHeight(int(screen_height * 0.8))
        self.adjustSize()

    @staticmethod
    def open_url(url: str):
        Popen(["xdg-open", url], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL, stdin=subprocess.DEVNULL)

    def _show_full_field_val(self, val: str):
        self.gbox_info.hide()
        self.bt_back.setVisible(True)
        self.text_field.show()
        self.text_field.setPlainText(val)

    def _gen_show_button(self, idx: int, val: str):

        is_url = self._can_open_url and bool(RE_URL.match(val)) if val else False

        if is_url:
            bt_label = self.i18n["manage_window.info.open_url"]

            def _show_field():
                self.open_url(val)
        else:
            bt_label = self.i18n["show"].capitalize()

            def _show_field():
                self._show_full_field_val(val)

        bt_full_field = QPushButton(bt_label)
        bt_full_field.setObjectName("show")
        bt_full_field.setCursor(QCursor(Qt.PointingHandCursor))
        bt_full_field.clicked.connect(_show_field)
        self.gbox_info.layout().addWidget(bt_full_field, idx, 2)

    def back_to_info(self):
        self.text_field.setPlainText("")
        self.text_field.hide()
        self.gbox_info.show()
        self.bt_back.setVisible(False)



from typing import List, Dict, Any, NamedTuple, Optional, Union, Collection, Iterable

from bauh.api.abstract.model import SoftwarePackage
from bauh.view.qt.view_model import PackageView


class PackageFilters(NamedTuple):
    """
    It represents the manage window selected filters
    """

    display_limit: int
    category: str
    name: Optional[str]
    only_apps: bool
    only_installed: bool
    only_updates: bool
    only_verified: bool
    search: Optional[str]  # initial search term
    type: str

    @property
    def anything(self) -> bool:
        return not self.only_installed and not self.only_updates and not self.only_apps \
            and not self.only_verified and not self.name and self.type == "any" and self.category == "any"


def new_pkgs_info() -> Dict[str, Any]:
    return {'apps_count': 0,  # number of application packages
            'napps_count': 0,  # number of not application packages (libraries, runtimes or something else)
            'available_types': {},  # available package types in 'new_pkgs'
            'updates': 0,
            'app_updates': 0,
            'napp_updates': 0,
            'pkgs_displayed': [],
            'not_installed': 0,
            'installed': 0,
            'categories': set(),
            'verified': 0,
            'pkgs': []}  # total packages


def update_info(pkgv: PackageView, pkgs_info: Dict[str, Any]):
    pkgs_info['available_types'][pkgv.model.get_type()] = {'icon': pkgv.model.get_type_icon_path(), 'label': pkgv.get_type_label()}

    if pkgv.model.is_application():
        pkgs_info['apps_count'] += 1
    else:
        pkgs_info['napps_count'] += 1

    if pkgv.model.is_trustable():
        pkgs_info['verified'] += 1

    if pkgv.model.update and not pkgv.model.is_update_ignored():
        if pkgv.model.is_application():
            pkgs_info['app_updates'] += 1
        else:
            pkgs_info['napp_updates'] += 1

        pkgs_info['updates'] += 1

    if pkgv.model.categories:
        for c in pkgv.model.categories:
            if c:
                cat = c.lower().strip()
                if cat:
                    pkgs_info['categories'].add(cat)

    pkgs_info['pkgs'].append(pkgv)

    if pkgv.model.installed:
        pkgs_info['installed'] += 1
    else:
        pkgs_info['not_installed'] += 1


def apply_filters(pkg: PackageView, filters: PackageFilters, info: dict, limit: bool = True):
    if not limit or not filters.display_limit or len(info['pkgs_displayed']) < filters.display_limit:
        if not is_package_hidden(pkg, filters):
            info['pkgs_displayed'].append(pkg)


def sum_updates_displayed(info: dict) -> int:
    updates = 0
    if info['pkgs_displayed']:
        for p in info['pkgs_displayed']:
            if p.model.update and not p.model.is_update_ignored():
                updates += 1

    return updates


def is_package_hidden(pkg: PackageView, filters: PackageFilters) -> bool:
    hidden = filters.only_installed and not pkg.model.installed

    if not hidden and filters.only_apps:
        hidden = pkg.model.installed and not pkg.model.is_application()

    if not hidden and filters.only_updates:
        hidden = not pkg.model.update or pkg.model.is_update_ignored()

    if not hidden and filters.only_verified:
        hidden = not pkg.model.is_trustable()

    if not hidden and filters.type is not None and filters.type != "any":
        hidden = pkg.model.get_type() != filters.type

    if not hidden and filters.category is not None and filters.category != "any":
        hidden = not pkg.model.categories or not next((c for c in pkg.model.categories if c.lower() == filters.category), None)

    if not hidden and filters.name:
        hidden = not filters.name.startswith(pkg.model.name.lower())

    return hidden


def _by_name(pkg: Union[SoftwarePackage, PackageView]):
    return pkg.name.lower()


def sort_packages(pkgs: Iterable[Union[SoftwarePackage, PackageView]], word: Optional[str],
                  limit: int = 0) -> List[SoftwarePackage]:
    exact, starts_with, contains, others = [], [], [], []

    if not word:
        others.extend(pkgs)
    else:
        for p in pkgs:
            lower_name = p.name.lower()
            if word == lower_name:
                exact.append(p)
            elif lower_name.startswith(word):
                starts_with.append(p)
            elif word in lower_name:
                contains.append(p)
            else:
                others.append(p)

    res = []
    for app_list in (exact, starts_with, contains, others):
        if app_list:
            last = limit - len(res) if limit is not None and limit > 0 else None

            if last is not None and last <= 0:
                break

            to_add = app_list[0:last]
            to_add.sort(key=_by_name)
            res.extend(to_add)

    return res
